// +-------------------------------- DISCLAIMER ---------------------------------+
// |                                                                             |
// | This application program is provided to you free of charge as an example.   |
// | Despite the considerable efforts of Euresys personnel to create a usable    |
// | example, you should not assume that this program is error-free or suitable  |
// | for any purpose whatsoever.                                                 |
// |                                                                             |
// | EURESYS does not give any representation, warranty or undertaking that this |
// | program is free of any defect or error or suitable for any purpose. EURESYS |
// | shall not be liable, in contract, in torts or otherwise, for any damages,   |
// | loss, costs, expenses or other claims for compensation, including those     |
// | asserted by third parties, arising out of or in connection with the use of  |
// | this program.                                                               |
// |                                                                             |
// +-----------------------------------------------------------------------------+
//
// ImageProcessing.cs: implementation of the ImageProcessing class.
//
//////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;

/// Open eVision Namespace
using Euresys.Open_eVision_1_2;


namespace ClinicalAnalysis
{


    class ImageProcessing
    {
        /// Source and temporary images
        private EImageBW8 m_SrcImg = new EImageBW8();
        private EImageBW8 m_ImgProcess = new EImageBW8();

        /// Source and temporary masks
        private EImageBW8 m_MaskImg = new EImageBW8();
        private EImageBW8 m_MaskProcess = new EImageBW8();

        /// ROI to analyze one test tube at a time
        EROIBW8 m_SrcROI = new EROIBW8();
        EROIBW8 m_MaskROI = new EROIBW8();

        /// EasyGauge tools to register the mask
        private EWorldShape m_WorldShape = new EWorldShape();
        private ELineGauge m_LineGauge = null;
        private EPointGauge m_PointGauge = null;

        /// EasyObject tools to detect reactions
        private ECodedImage2[] m_CodedImage = new ECodedImage2[4];
        private EImageEncoder m_Encoder = new EImageEncoder();
        private EObjectSelection[] m_Selection = new EObjectSelection[4];

        /// Canny Edge detector tool to detect bubbles
        private ECannyEdgeDetector m_CannyEdge = new ECannyEdgeDetector();

        /// Reference offset and rotation values
        private float m_ReferenceAngle = 0.0f;
        private float m_ReferencePositionX = 0.0f;
        private float m_ReferencePositionY = 0.0f;

        /// <summary>
        /// Constructor
        /// </summary>
        public ImageProcessing()
        {

        }


        /// <summary>
        /// Destructor
        /// </summary>
        ~ImageProcessing()
        {

        }


        /// <summary>
        /// Initialization
        /// </summary>
        public void Initialize()
        {

            for (int i = 0; i < 4; i++)
            {
                /// Initialize the EasyObject tools
                m_CodedImage[i] = new ECodedImage2();
                m_Selection[i] = new EObjectSelection();
            }

            try
            {
                /// Load the EasyGauge model file 
                m_WorldShape.Load(@"..\..\..\Resources\Model\ClinicalAnalysis.GGE", true);
                m_LineGauge = (ELineGauge)m_WorldShape.GetDaughter(0);
                m_PointGauge = (EPointGauge)m_LineGauge.GetDaughter(0);

                /// Load the reference image 
                /// This reference image was used to create the flexible mask
                m_SrcImg.Load(@"..\..\..\Resources\Images\ClinicalAnalysis_01.bmp");

                /// Evaluate the reference image offset and rotation
                m_WorldShape.Process(m_SrcImg);
                m_ReferenceAngle = m_LineGauge.MeasuredLine.Angle;
                m_ReferencePositionX = m_LineGauge.MeasuredLine.CenterX + m_PointGauge.GetMeasuredPoint().X;
                m_ReferencePositionY = m_LineGauge.MeasuredLine.CenterY + m_PointGauge.GetMeasuredPoint().Y;

                /// Load the corresponding flexible mask 
                m_MaskImg.Load(@"..\..\..\Resources\Mask\FlexMask.bmp");

                /// Resize the temporary mask image
                m_MaskProcess.SetSize(m_MaskImg);
                
                /// Attach an ROI to it
                m_MaskROI.Attach(m_MaskProcess); 
                
                // Empty the source image object
                m_SrcImg.SetSize(0, 0);
          }
          catch (EException exc)
          {
              MessageBox.Show(exc.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
          }

        }


        /// <summary>
        /// Drawing
        /// </summary>
        public void Draw(Graphics g)
        {

            // Specify the color for drawing
            ERGBColor redColor = new ERGBColor(255, 0, 0); 
            ERGBColor blueColor = new ERGBColor(0, 0, 255); 

            try
            {
                /// Exit if no image has been loaded
                if (m_SrcImg.IsVoid)
                    return;

                /// Draw the source image
                m_SrcImg.Draw(g);


                /// Draw the detected objects for each reagents 
                for (int i = 0; i < 4; i++)
                {
                    if (m_Selection[i].ElementCount > 0)
                    {
                        /// Draw the detected objects
                        m_CodedImage[i].Draw(g, redColor, m_Selection[i]);                       
                        
                        /// Draw the objects gravity center
                        m_CodedImage[i].DrawFeature(g, blueColor, EDrawableFeature.GravityCenter, m_Selection[i]);
                    }
                }
            }
            catch (EException exc)
            {
                MessageBox.Show(exc.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }


        }

        /// <summary>
        /// Inspection.
        /// </summary>
        public void Inspect(string filename)
        {
            /// Temporary image for preprocessing
            EImageBW8 m_TempImg = new EImageBW8();
            try
            {
                /// Load the image for inspection
                m_SrcImg.Load(filename);

                /// Resize the temporary image accordingly
                m_ImgProcess.SetSize(m_SrcImg);

                /// Attach an ROI to it
                m_SrcROI.Attach(m_ImgProcess);

                /// Evaluate the offset and the rotation for the current image 
                m_WorldShape.Process(m_SrcImg, true);
                float m_Angle = m_LineGauge.MeasuredLine.Angle - m_ReferenceAngle;
                float m_X1 = m_LineGauge.MeasuredLine.CenterX + m_PointGauge.GetMeasuredPoint().X;
                float m_Y1 = m_LineGauge.MeasuredLine.CenterY + m_PointGauge.GetMeasuredPoint().Y;

                /// Register the mask according to the offset and the rotation of the current image
                EasyImage.ScaleRotate(m_MaskImg, m_ReferencePositionX, m_ReferencePositionY, m_X1, m_Y1, 1.000f, 1.000f, -m_Angle, m_MaskProcess, 8);

                for (int i = 0; i < 4; i++)
                {

                    /// Make a copy of the image for processing
                    EasyImage.Oper(EArithmeticLogicOperation.Copy, m_SrcImg, m_ImgProcess);

                    /// Position the image ROI
                    m_SrcROI.SetPlacement((int)(m_X1) - 25 - (150 * (2 - i)), 2, 150, 308);

                    /// Position the mask ROI
                    m_MaskROI.SetPlacement((int)(m_X1) - 25 - (150 * (2 - i)), 2, 150, 308);

                    /// Clear the objects selection
                    m_Selection[i].Clear();

                    /// Set the thresholding mode for the blob analysis and specify which layers to encode
                    m_Encoder.GrayscaleSingleThresholdSegmenter.Mode = EGrayscaleSingleThreshold.Absolute;
                    m_Encoder.GrayscaleSingleThresholdSegmenter.BlackLayerEncoded = true;
                    m_Encoder.GrayscaleSingleThresholdSegmenter.WhiteLayerEncoded = false;

                    /// Process each Reagents Separately 
                    switch (i)
                    {
                        /// Reagent 1
                        case 0:
                            /// Preprocess the image for bubble detection
                            m_TempImg.SetSize(m_ImgProcess);
                            EasyImage.WhiteTopHatDisk(m_ImgProcess, m_TempImg, 5);
                            EasyImage.Median(m_TempImg, m_ImgProcess);

                            /// Canny edge for bubbles detection
                            m_CannyEdge.HighThreshold = 0.2f;
                            m_CannyEdge.LowThreshold = 0.1f;
                            m_CannyEdge.ThresholdingMode = ECannyThresholdingMode.Relative;
                            m_CannyEdge.SmoothingScale = 1.100f;
                            m_CannyEdge.Apply(m_ImgProcess, m_ImgProcess);

                            /// Set the absolute threshold value and specify which layers to encode
                            m_Encoder.GrayscaleSingleThresholdSegmenter.AbsoluteThreshold = 90;
                            m_Encoder.GrayscaleSingleThresholdSegmenter.BlackLayerEncoded = false;
                            m_Encoder.GrayscaleSingleThresholdSegmenter.WhiteLayerEncoded = true;

                            /// Encode the objects with masking
                            m_Encoder.Encode(m_SrcROI, m_MaskROI, m_CodedImage[i]);
                            m_Selection[i].AddObjects(m_CodedImage[i]);
                            m_Selection[i].RemoveUsingUnsignedIntegerFeature(EFeature.Area, 20, ESingleThresholdMode.LessEqual);
                            break;

                        /// Reagent 2
                        case 1:
                            /// Set the absolute threshold value
                            m_Encoder.GrayscaleSingleThresholdSegmenter.AbsoluteThreshold = 80;
                            /// Encode objects with masking
                            m_Encoder.Encode(m_SrcROI, m_MaskROI, m_CodedImage[i]);
                            m_Selection[i].AddObjects(m_CodedImage[i]);
                            m_Selection[i].RemoveUsingUnsignedIntegerFeature(EFeature.Area, 50, ESingleThresholdMode.LessEqual);
                            break;

                        /// Reagent 3
                        case 2:
                            /// Set the absolute threshold value
                            m_Encoder.GrayscaleSingleThresholdSegmenter.AbsoluteThreshold = 80;
                            /// Encode objects with masking
                            m_Encoder.Encode(m_SrcROI, m_MaskROI, m_CodedImage[i]);
                            m_Selection[i].AddObjects(m_CodedImage[i]);
                            break;

                        /// Reagent 4
                        case 3:
                            /// Set the absolute threshold value
                            m_Encoder.GrayscaleSingleThresholdSegmenter.AbsoluteThreshold = 50;
                            /// Encode objects with masking
                            m_Encoder.Encode(m_SrcROI, m_MaskROI, m_CodedImage[i]);
                            m_Selection[i].AddObjects(m_CodedImage[i]);
                            m_Selection[i].RemoveUsingUnsignedIntegerFeature(EFeature.Area, 100, ESingleThresholdMode.LessEqual);
                            break;

                    }

                }

            }
            catch (EException exc)
            {
                MessageBox.Show(exc.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }


        }

    }
}


